﻿// --------------------------------------------------------------------------------
// File: ApgEmeaUsbDrawer.cs
//
// Copyright © 2019 APG Cash Drawer Ltd. 
//
// No part of this document may be reproduced or transmitted 
// by any form, or in any means including photocopying, and 
// recording, without the prior permission of APG Cash Drawer Ltd.
//
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
// OR OTHER DEALINGS IN THE SOFTWARE.
//-------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.IO.Ports;
using System.ComponentModel;            
using System.Diagnostics;               //Added for debuggin purposes in this sample

namespace APG_EMEA_USB_Sample
{
    /// <summary>
    /// This is a partial implementention of the APG EMEA USB Drawer bussines class
    /// The class provides a very basic management of serial communications via VCP
    /// It allows to open/close a serial connection; Send a firing open command
    /// and receive drawer status. It does not manage exceptions and serves only
    /// as a very basis sample on how to interface with a USB Drawer.
    /// </summary>
    class ApgEmeaUsbDrawer
    {
        #region Constants
        /// <summary>
        /// Serial port configuration constants
        /// </summary>
        const int BAUD_RATE         = 9600;
        const int DATA_BITS         = 8;
        const StopBits STOP_BITS    = StopBits.One;
        const Parity PARITY         = Parity.None;

        /// <summary>
        /// USB Drawer HEX Codes
        /// </summary>
        const string FIRING_CODE            = "07";                         //HEX Code to request drawer opening
        const string GET_STATUS_CODE        = "1D-61";                      //HEX Code to request drawer status
        const string STATUS_OPEN_CODE       = "10-00-00-0F";                //HEX Code sent by drawer to notify drawer status changed to open
        const string STATUS_CLOSE_CODE      = "14-00-00-0F";                //HEX Code sent by drawer to notify drawer status changed to closed
        const string NAK_CODE               = "15-15-15-15";                //HEX Code sent by drawer when a command is not recognized
        const string MEMORY_OVERFLOW_CODE   = "4F-76-65-72-66-6C-6F-77";    //HEX Code sent by drawer to notify counts memory issue (Overflow).
        const string END_OF_LINE            = "0D-0A";                      //HEX Codes to identify the end of a response on NEW USB drawer commands: '/n/r'
        #endregion

        #region Public Properties
        public bool? IsOpen { get; private set; }
        public bool IsConnected { get; private set; }
        #endregion

        #region Private Properties
        private SerialPort _comPort;                                        //Object for managing serial COM port communication
        private BackgroundWorker _bgwComListener = new BackgroundWorker();  //Background worker for listening incomming messages
        private List<string> _commandList;                                  //List used to check if the received message is a USB command
        #endregion

        #region Events
        public delegate void ChangeEventHandler(object sender, EventArgs e); //Handler for general events (no arguments)

        public event ChangeEventHandler DrawerStatusChanged;                //Open/Closed status
        public event ChangeEventHandler DrawerConnectionStatusChanged;      //Connected/Disconnected status
        #endregion

        #region Constructor
        /// <summary>
        /// Default constructor
        /// </summary>
        public ApgEmeaUsbDrawer()
        {
            //Populates the command list. This list is used to check if an actual command has been received.
            _commandList = new List<string>{
                FIRING_CODE,
                GET_STATUS_CODE,
                STATUS_OPEN_CODE,
                STATUS_CLOSE_CODE,
                NAK_CODE,
                MEMORY_OVERFLOW_CODE,
            };

            //Set the default status
            IsOpen = null;
            IsConnected = false;

            //Configure background worker - Serial Listener
            _bgwComListener.WorkerSupportsCancellation = true;
            _bgwComListener.DoWork += _bgwComListener_DoWork;
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Open a serial communication to the provided COM port
        /// </summary>
        /// <param name="strComPort"></param>
        public async void Connect(string strComPort)
        {
            //Try to connect to COM Port
            bool isConnected = await _connect(strComPort);
            if(isConnected)
            {
                //Start background worker
                if(!_bgwComListener.IsBusy)
                    _bgwComListener.RunWorkerAsync();

                _sendCommand(GET_STATUS_CODE); //Send get status to identify current drawer status
            }
        }

        /// <summary>
        /// Closes the serial communication 
        /// </summary>
        public void Disconnect()
        {
            _disconnect();
        }

        /// <summary>
        /// Sends OPEN command to USB drawer
        /// </summary>
        public void Open()
        {
            _sendCommand(FIRING_CODE); //Convert firing code to byte array

        }

        /// <summary>
        /// Sends GET STATUS command to USB drawer
        /// </summary>
        public async void GetStatus()
        {
            _sendCommand(GET_STATUS_CODE);
        }

        /// <summary>
        /// Clear the drawer status property setting it to UNKNOWN status
        /// </summary>
        public void ClearStatus()
        {
            IsOpen = null;
            DrawerStatusChanged?.Invoke(this, new EventArgs()); //Raise event
        }
        #endregion

        #region Private Methods
        private async Task<bool> _connect(string strComPort)
        {
            if(_comPort == null)
            {
                _comPort = new SerialPort();
            }

            if(!_comPort.IsOpen)
            {
                try
                {
                    //Set Serial Port parameters
                    _comPort.PortName = strComPort;
                    _comPort.BaudRate = BAUD_RATE;
                    _comPort.DataBits = DATA_BITS;
                    _comPort.Parity = PARITY;

                    _comPort.Open();
                    IsConnected = true;
                    DrawerConnectionStatusChanged?.Invoke(this, new EventArgs()); //Raise event

                    return true;
                }
                catch (Exception e)
                {
                    Debug.WriteLine("*** ERROR: opening device (COM Port: " + strComPort + "). " + e.Message);
                }
            }
            else
            {
                Debug.WriteLine("*** ERROR: opening device (COM Port: " + strComPort + "). Port was already opened.");
            }
            
            return false;
        }

        private async Task<bool> _disconnect()
        {
            if (_comPort == null)
            {
                IsConnected = false;
                return false;
            }

            if (_bgwComListener.IsBusy)
            {
                _bgwComListener.CancelAsync();
                _sendCommand(GET_STATUS_CODE);
                Thread.Sleep(500);
            }

            _comPort.Close();
            IsConnected = false;
            IsOpen = null;
            DrawerStatusChanged?.Invoke(this, new EventArgs()); //Raise event
            DrawerConnectionStatusChanged?.Invoke(this, new EventArgs()); //Raise event

            Debug.WriteLine("Connection closed.");

            return true;
        }

        /// <summary>
        /// Sends a command to the USB drawer
        /// </summary>
        /// <param name="Command">The command must be provided in HEX format</param>
        /// <returns></returns>
        private async Task _sendCommand(string Command)
        {
            byte[] bCommand = _hexToBytes(Command); //Convert command code to byte array

            if (IsConnected != true)
                throw new Exception("The drawer is not connected.");

            try
            {
                _comPort.Write(bCommand, 0, bCommand.Length); //Send firing code
                Debug.WriteLine("Write Command: " + BitConverter.ToString(bCommand.ToArray()) + "; NumBytesWritten: " + bCommand.Length);
            }
            catch(Exception e)
            {
                Debug.WriteLine("*** ERROR: writing to COM port. " + e.Message);
            }
        }

        /// <summary>
        /// Read a USB message from the COM buffer. 
        /// Function reads one byte at a time from the serial port buffer. USB commands do not have
        /// delimitators so we have to check every byte until a full message is recieved.
        /// </summary>
        /// <returns>Byte array with the received command.</returns>
        private byte[] _readCommand()
        {
            List<byte> tempMessage = new List<byte>();  //Array to keep temporary message;
            bool fIsEndOfLine = false;                  //Flag to mark a complete message received.

            do //Loop until a full message is received
            {
                try
                {
                    byte[] bReadByte = new byte[1];
                    _comPort.Read(bReadByte, 0, bReadByte.Length); //Read a byte from the serial buffer
                    tempMessage.AddRange(bReadByte); //Concatenate the read byte to the temporary byte array
                }
                catch (Exception e)
                {
                    Debug.WriteLine("*** ERROR: reading from COM port. " + e.Message);
                    _bgwComListener.CancelAsync(); //
                    _disconnect();
                    return tempMessage.ToArray();
                }

                if (_isEndOfLine(tempMessage.ToArray())) //If a full message has been received -> exit the loop.
                    fIsEndOfLine = true;

            } while (!fIsEndOfLine);

            Debug.WriteLine("Read Command: " + BitConverter.ToString(tempMessage.ToArray()) + "; NumBytesRead: " + tempMessage.Count);
            return tempMessage.ToArray();

        }

        /// <summary>
        /// Check if a byte array is a respond from a USB drawer.
        /// It checks if the array is one of the event responses from a USB drawer or if the return carriage character
        /// is in the message.
        /// </summary>
        /// <param name="tempBuffer">Byte array to be checked</param>
        /// <returns>True if </returns>
        private bool _isEndOfLine(byte[] tempBuffer)
        {
            //Convert bytes to HEX string
            string strMessageIn = BitConverter.ToString(tempBuffer.ToArray());

            //Check if the received message contains one of the USB responses
            if (_commandList.Contains(strMessageIn))
                return true;

            //If it is not a old USB standar response check if a return carriage character is in the message
            if (strMessageIn.Contains(END_OF_LINE))//Return line /r/n
                return true;

            return false;
        }

        /// <summary>
        /// Parses commands and 
        /// </summary>
        /// <param name="bMessageIn">Byte array with the received command</param>
        private void _parseMessage(byte[] bMessageIn)
        {
            //Convert bytes to HEX string
            string strMessageIn = BitConverter.ToString(bMessageIn.ToArray());

            if (strMessageIn == STATUS_OPEN_CODE)//Drawer Open
            {
                IsOpen = true;
                DrawerStatusChanged?.Invoke(this, new EventArgs()); //Raise event
            }

            if (strMessageIn == STATUS_CLOSE_CODE)//Drawer Closed
            {
                IsOpen = false;
                DrawerStatusChanged?.Invoke(this, new EventArgs()); //Raise event
            }

        }

        /// <summary>
        /// Converts a string in HEX format to a byte array. 
        /// Accepted formats are:   1D-4F
        ///                         1D 4F
        ///                         1D4F
        /// </summary>
        /// <param name="strHex">String to be converted</param>
        /// <returns>Byte array with the sting converted</returns>
        private static byte[] _hexToBytes(string strHex)
        {
            strHex.Trim();
            strHex = strHex.Replace("-", string.Empty);
            strHex = strHex.Replace(" ", string.Empty);

            byte[] btString = new byte[strHex.Length / 2];

            for (int i = 0; i < btString.Length * 2; i += 2)
            {
                btString[i / 2] = Convert.ToByte(strHex.Substring(i, 2), 16);
            }
            return btString;
        }

        /// <summary>
        /// Background Worker that keeps listening to COM port for drawer events such as
        /// change of status (drawer open / closed)
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void _bgwComListener_DoWork(object sender, DoWorkEventArgs e)
        {
            byte[] bMessageIn;

            do //Infinite loop reading commands
            {
                //Check if there is a cancellation request pending for the background worker.
                if (_bgwComListener.CancellationPending)
                {
                    e.Cancel = true;
                    return;
                }

                bMessageIn = _readCommand(); //Read next message (this function blocks the thread if no byte is received)
                _parseMessage(bMessageIn);
            } while (true);
        }
        #endregion

    }
}
